<?php

namespace Drupal\mustache_magic\Plugin\mustache\Magic;

use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Render\BubbleableMetadata;
use Drupal\Core\Render\RenderContext;
use Drupal\Core\Url;
use Drupal\mustache\Helpers\MustacheRenderTemplate;
use Drupal\mustache\Plugin\MustacheMagic;
use Drupal\mustache\Render\IterableMarkup;
use Drupal\mustache\Render\Markup;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Enable automated DOM content synchronization by using a lambda.
 *
 * Usage:
 *
 * You can use the "magical" {{#sync.<options>}} variable to enable automated
 * DOM content synchronization. Examples:
 *
 * @code
 * One time immediate refresh: {{#sync.now}}{{node.title}}{{/sync.now}}
 * Unlimited refresh every 10 seconds: {{#sync.now.10.always}}...{{/sync.now.10.always}}
 * Up to 10 refreshes every second: {{#sync.now.1.10}}...{{/sync.now.1.10}}
 * Up to 10 refreshes every second, delayed by 5 seconds: {{#sync.5.1.10}}...{{/sync.5.1.10}}
 * Trigger refresh when clicking a DOM element having the CSS class ".button": {{#sync.trigger.button.click}}...{{/sync.trigger.button.click}}
 * @endcode
 *
 * @see https://git.drupalcode.org/project/mustache_templates/-/blob/2.0.x/README.md#36-magic-synchronization
 *
 * @MustacheMagic(
 *   id = "sync",
 *   label = @Translation("Synchronization"),
 *   description = @Translation("Use the <b>{{#sync.&lt;options&gt;}}</b> variable to enable automated DOM content synchronization. Examples: <ul><li>One time immediate refresh: {{#sync.now}}{{node.title}}{{/sync.now}}</li><li>Unlimited refresh every 10 seconds: {{#sync.10.always}}...{{/sync.10.always}}</li><li>Up to 10 refreshes every second: {{#sync.1.10}}...{{/sync.1.10}}</li><li>Up to 10 refreshes every second, delayed by 5 seconds: {{#sync.5.1.10}}...{{/sync.5.1.10}}</li><li>Trigger refresh when clicking a DOM element having the CSS class .button: {{#sync.trigger.button.click}}...{{/sync.trigger.button.click}}</li></ul>More examples can be found in the <a href='https://git.drupalcode.org/project/mustache_templates/-/blob/2.0.x/README.md#36-magic-synchronization' target='_blank' rel='noopener noreferrer'>README</a>.")
 * )
 */
class Sync extends MustacheMagic {

  /**
   * A sequence of user-defined synchronization settings.
   *
   * @var array
   */
  protected $keys = [];

  /**
   * Whether this object is a clone.
   *
   * @var bool
   */
  protected $cloned = FALSE;

  /**
   * The renderer.
   *
   * @var \Drupal\Core\Render\RendererInterface
   */
  protected $renderer;

  /**
   * The sync storage.
   *
   * @var \Drupal\mustache_magic\MustacheSyncStorage
   */
  protected $syncStorage;

  /**
   * The URL to use for server-side synchronization of user-defined templates.
   *
   * @var string
   */
  protected static $syncUrl = '/m/sync?';

  /**
   * A list of keywords to exclude when a certain scope is being handled.
   *
   * @var array
   */
  protected static $excludes = [
    'now',
    'always',
    'once',
    'span',
    'form',
    'trigger',
    'proxy',
    'increment',
  ];

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    $instance = parent::create($container, $configuration, $plugin_id, $plugin_definition);
    $instance->renderer = $container->get('renderer');
    $instance->syncStorage = $container->get('mustache.sync_storage');
    return $instance;
  }

  /**
   * Implementation of the magic __isset() method.
   */
  public function __isset($name) {
    return !empty($name);
  }

  /**
   * Implementation of the magic __get() method.
   */
  public function __get($name) {
    if (!$this->cloned) {
      $cloned = clone $this;
      $cloned->cloned = TRUE;
      return $cloned->__get($name);
    }
    $name = mb_strtolower(trim($name));
    if (ctype_digit(strval($name))) {
      $this->keys[] = (int) $name;
    }
    else {
      $this->keys[] = $name;
    }
    return $this;
  }

  /**
   * Implements the magic __invoke() method, used as higher-order section.
   */
  public function __invoke($template_content, $render) {
    $template_hash = hash('md4', $template_content);
    $element_sync = !empty($this->element['#use_sync']) && isset($this->element['#sync']['items']) ? reset($this->element['#sync']['items']) : NULL;
    $bubbleable_metadata = BubbleableMetadata::createFromRenderArray($this->element);

    $build = MustacheRenderTemplate::build($template_hash, $template_content);
    $sync_options = $build->withClientSynchronization()->morphing();
    $build_render_array = &$build->toRenderArray();

    $to_merge = [];
    $to_copys = [
      '#cache',
      '#partials',
      '#inline_partials',
    ];
    foreach ($to_copys as $to_copy) {
      if (!empty($this->element[$to_copy])) {
        $to_merge[$to_copy] = $this->element[$to_copy];
      }
    }
    if ($element_sync) {
      $to_merge['#sync'] = [];
      $to_copys = ['eval', 'behaviors'];
      foreach ($to_copys as $to_copy) {
        if (!empty($element_sync[$to_copy])) {
          $to_merge['#sync'][$to_copy] = $element_sync[$to_copy];
        }
      }
    }

    $force_proxy = FALSE;
    foreach ($this->keys as $i => $key) {
      if ($i === 0) {
        if (is_int($key) && $key > 0) {
          $sync_options->startsDelayed($key * 1000);
          continue;
        }
        elseif ($key === 'now') {
          continue;
        }
      }
      if ($key === 'always') {
        $sync_options->unlimited();
        continue;
      }
      if ($key === 'span') {
        $sync_options->withWrapperTag('span');
        continue;
      }
      if ($key === 'proxy') {
        $force_proxy = TRUE;
        continue;
      }
      if (isset($form_bind) && $form_bind === 0) {
        $form_bind = 1;
        if (!is_int($key) && !in_array($key, static::$excludes)) {
          $form_selector = $key;
          if ($form_selector !== 'form'
            && !in_array(mb_substr(
            $form_selector, 0, 1), ['#', '[', '*', ':', '.'])) {
            $form_selector = '.' . $form_selector;
          }
          continue;
        }
      }
      if ($key === 'form') {
        $form_bind = 0;
        $form_selector = 'form';
        continue;
      }
      if (isset($trigger) && $trigger < 3) {
        $trigger++;
        if ($trigger === 1) {
          $trigger_selector = $key;
          if (!in_array(mb_substr(
            $trigger_selector, 0, 1), ['#', '[', '*', ':', '.'])) {
            $trigger_selector = '.' . $trigger_selector;
          }
        }
        elseif ($trigger === 2) {
          $trigger_event = !is_int($key) && !in_array($key, static::$excludes) ? $key : 'click';
        }
        elseif ($trigger === 3) {
          if ($key == 'once') {
            $trigger_limit = 1;
          }
          elseif (is_int($key) && $key > 0) {
            $trigger_limit = $key;
          }
          else {
            $trigger_limit = -1;
          }
        }
        continue;
      }
      if ($key === 'trigger' && !isset($trigger)) {
        $trigger = 0;
        $trigger_selector = '*';
        $trigger_event = 'click';
        $trigger_limit = -1;
        continue;
      }
      if (is_int($key)) {
        if (!isset($period)) {
          $period = $key * 1000;
          $sync_options->periodicallyRefreshesAt($period);
        }
        else {
          $sync_options->upToNTimes($key);
        }
        continue;
      }
      if (isset($url) && $url === '') {
        $url = str_replace('_dot_', '.', $key);
        if (substr($url, 0, 1) !== '/' && substr($url, 0, 4) !== 'http') {
          $url = '/' . $url;
        }
        continue;
      }
      if ($key === 'url') {
        $url = '';
        continue;
      }
      if (isset($increment) && $increment < 4) {
        $increment++;
        if ($increment === 1) {
          $increment_key = $key;
        }
        elseif ($increment === 2) {
          $increment_offset = $key;
        }
        elseif ($increment === 3) {
          if ($key == 'once') {
            $increment_limit = 1;
          }
          elseif (is_int($key) && $key > 0) {
            $increment_limit = $key;
          }
          else {
            $increment_limit = -1;
          }
        }
        elseif ($increment === 4) {
          $increment_step_size = $key;
        }
        continue;
      }
      if ($key === 'increment' || $key === 'increments') {
        $increment = 0;
        $increment_key = 'page';
        $increment_offset = 0;
        $increment_limit = -1;
        $increment_step_size = 1;
        continue;
      }
    }
    if (isset($form_bind)) {
      $sync_options->usingFormValues($form_selector);
    }
    if (isset($trigger)) {
      $sync_options->startsWhenElementWasTriggered($trigger_selector)
        ->atEvent($trigger_event)
        ->upToNTimes($trigger_limit);
    }
    if (isset($increment)) {
      $sync_options->increments()
        ->atParamKey($increment_key)
        ->startingAt($increment_offset)
        ->upToNTimes($increment_limit)
        ->withStepSize($increment_step_size);
    }

    if (empty($url) && $element_sync && !empty($element_sync['url'])) {
      $url = $element_sync['url'];
    }
    if (!empty($this->element['#data'])) {
      if (is_array($this->element['#data'])) {
        $data = $this->element['#data'];
      }
      elseif (empty($url) && (is_string($this->element['#data']) || $this->element['#data'] instanceof Url)) {
        $url = $this->element['#data'];
      }
    }
    if (!isset($data)) {
      if (isset($this->element['#override_data']) && is_array($this->element['#override_data'])) {
        $data = $this->element['#override_data'];
      }
      else {
        $data = [];
      }
    }
    foreach ($data as $d_k => $d_v) {
      if ($d_v instanceof IterableMarkup) {
        unset($data[$d_k]);
      }
    }

    if (empty($url)) {
      $build->withPlaceholder(['#markup' => Markup::create($render($template_content))]);
    }
    elseif (empty($data)) {
      $build->usingDataFromUrl($url);
    }
    else {
      $build->usingData($data);
      $sync_options->usingDataFromUrl($url);
    }

    if (empty($url) || $force_proxy) {
      $values = [
        'name' => $template_hash,
        'content' => $template_content,
        'merge' => $to_merge,
      ];

      $server_side_render = $build->toRenderArray();
      $server_side_render['#sync'] = [];

      if (!empty($this->element['#with_tokens'])) {
        $server_side_render['#with_tokens'] = $this->element['#with_tokens'];
        /** @var \Drupal\mustache\MustacheTokenProcessor $token_processor */
        $token_processor = \Drupal::service('mustache.token_processor');
        $values['token_options'] = !empty($this->element['#with_tokens']['options']) ? $this->element['#with_tokens']['options'] : [];
        $token_data = !empty($this->element['#with_tokens']['data']) ? $this->element['#with_tokens']['data'] : [];
        $token_processor->processData($token_data, 'view', $bubbleable_metadata);
      }

      $build->withPlaceholder($server_side_render);
      unset($build_render_array['#inline']);
      $build_render_array['#template'] = 'mustache_magic_sync';

      if ($force_proxy && !empty($url)) {
        if ($url instanceof Url) {
          $url = clone $url;
          $url->setAbsolute($url->isExternal());
          $url = $url->toString();
        }
        $values['url'] = $url;
      }

      $datas = isset($token_data) ? [
        'data' => $data,
        'token_data' => $token_data,
      ] : ['data' => $data];
      foreach ($datas as $d_k => $d_v) {
        $values[$d_k] = [];
        foreach ($d_v as $t_key => $t_value) {
          if ($t_value instanceof EntityInterface) {
            if (!$t_value->isNew()) {
              $values[$d_k][$t_key] = [
                $t_value->getEntityTypeId(),
                $t_value->language()->getId(),
                $t_value->id(),
              ];
            }
            elseif ($uuid = $t_value->uuid()) {
              $values[$d_k][$t_key] = [$t_value->getEntityTypeId(), $uuid];
            }
          }
          elseif (is_scalar($t_value) || $t_value instanceof \JsonSerializable) {
            $values[$d_k][$t_key] = $t_value;
          }
          else {
            $values[$d_k][$t_key] = NULL;
          }
        }
      }

      $sync_hash = $this->syncStorage->generateHash($values);
      if (!$this->syncStorage->has($sync_hash)) {
        if ($sync_hash != $this->syncStorage->set($values)) {
          throw new \LogicException("The Mustache sync storage is behaving unexpected: Received a different hash for given values other than previously generated.");
        }
      }

      $sync_options->usingDataFromUrl(static::$syncUrl . 'h=' . urlencode($sync_hash));
    }
    else {
      foreach ($to_merge as $m_key => $m_val) {
        if ($m_key === '#sync') {
          $sync_render_array = &$sync_options->toRenderArray();
          $sync_render_array = NestedArray::mergeDeep($sync_render_array, $to_merge['#sync']);
        }
        elseif (isset($build_render_array[$m_key]) && is_array($build_render_array[$m_key])) {
          $build_render_array[$m_key] = NestedArray::mergeDeep($build_render_array[$m_key], $m_val);
        }
        else {
          $build_render_array[$m_key] = $m_val;
        }
      }
    }

    $renderer = $this->renderer;
    $rendered = $renderer->executeInRenderContext(new RenderContext(), function () use ($renderer, &$build_render_array) {
      return $renderer->render($build_render_array);
    });
    $bubbleable_metadata
      ->addCacheableDependency(BubbleableMetadata::createFromRenderArray($build_render_array))
      ->applyTo($this->element);
    // We need to replace the opening curly brackets, so that Mustache.php's
    // lambda helper cannot chime in and would try to replace Mustache
    // variables within the inline template with  current context values.
    return Markup::create(str_replace('{{', '\{\{', (string) $rendered));
  }

}
